#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from datetime import datetime
from collections import namedtuple
import struct
from enum import IntEnum
from .BrFile import *

BrProfilerEvent = namedtuple('BrProfilerEvent', [ 'microseconds', 'objectIdent', 'objectName', 'event', 'eventDescription'])

class BrProfilerObjectType(IntEnum):
	UNKNOWN = 0
	LIBRARY_FUNCTION = 1
	TASK_CLASS = 2
	CYCLIC_TASK = 3
	SYSTEM_TASK = 4
	INTERRUPT_VECTOR = 5

BrLibraryFunctionEvents = {
	0x03000000 : 'library function started',
	0x03000001 : 'library function finished'
}


BrCyclicTaskEvents = {
	0x02000000 : 'cyclic task started',
	0x02000001 : 'cyclic task finished'
}

BrTaskClassEvents = {
	0x1e000000 : 'taskclass started',
	0x1e000001 : 'taskclass stopped',
	0x1e000001 : 'input scheduler finished',
	0x1e000002 : 'end of cyclic programs',
	0x1e000003 : 'output scheduler started',
	0x1e000004 : 'output scheduler finished'
}


BrSystemObjectEvents = {
	0x00191902 : '#1 is now running',
	0x00252502 : '#2 is now running',
	0x00373704 : '#3 is now running',
	0x00414102 : '#8 is now running',	
	0x00020200 : 'IO Scheduler is now running',
	0x00fcfc02 : 'TcIdleFiller is now running',
	0x01000000 : 'system task started',
	0x01000001 : 'system task stopped',
	0x00212102 : 'system task is now running',
	0x11000001 : 'system task deleted'
}

BrInterruptHandlerEvents = {
	0x01000000 : 'interrupt handler started',
	0x01000001 : 'interrupt handler finished'	
}

class BrProfilerDataFile(BrFile):
	'''
	class for a *.br file containing profiler data
	'''
	SIZEOF_RECORD_ENTRY = 16
	SIZEOF_LIBRARY_FUNCTION_ENTRY = 28
	SIZEOF_TASKCLASS_ENTRY = 70
	OFFSET_OBJECT_ENTRY = 0x442
	SIZEOF_OBJECT_ENTRY = 62
	SIZEOF_SYSTEMOBJECT_ENTRY = 70
	MIN_SIZE_INTERRUPT_VECTOR = 32

	def __init__(self, filename: str):
		super().__init__(filename)
		if self.fileType != BrFileType.PROFILER_DATA_OBJECT:
			raise TypeError(f'content is not a B&R profiler data module (Type is {self.fileType})')
		self.__offsetRecords = struct.unpack_from('<H', self._content, 0x2d)[0] + struct.unpack_from('<H', self._content, 0x2f)[0] # 2 x uint16
							
		self.__maxSize = struct.unpack_from('<I', self._content, 0xa4)[0] # uint32 maximal size of profiler data object
		self.__additionalDatasizePerEntry = struct.unpack_from('<I', self._content, 0xa8)[0] # uint32
		self.__bufferForCreatedTask = struct.unpack_from('<I', self._content, 0xac)[0] # uint32		
		self.__bufferForCreatedUserTask = struct.unpack_from('<I', self._content, 0xbc)[0] # uint32
		head = 0x116
		# optional library functions
		libraryFunctionsList = bytearray()
		while struct.unpack_from('<H', self._content, head)[0] != 0:
			c = self._content[head]
			libraryFunctionsList.append( self._content[head])
			head +=1
		if len(libraryFunctionsList) > 0:
			head +=1
			libraryFunctionsList = str(libraryFunctionsList, 'ascii').split('\0')
		self.__sgType = 'SG' + str(struct.unpack_from('<B', self._content, head+2)[0])
		self.__arversion = str(self._content[head+6:head+8], encoding = 'ascii') + '.' \
					+ str(self._content[head+8:head+10], encoding = 'ascii')
		self.__bridcode = hex(struct.unpack_from('<I', self._content, head+10)[0]) # uint32
		self.__firstRecordAtOffset = struct.unpack_from('<I', self._content, head+14)[0] # uint32
		self.__lastRecordAtOffset = struct.unpack_from('<I', self._content, head+18)[0] # uint32
		self.__tickFrequencyMhz = struct.unpack_from('<I', self._content, head+22)[0] # uint32
		year, month, day, _ , hour, minute, second  = struct.unpack_from('<HBBBBBB', self._content, head+26)
		self.__profilingTimestamp = datetime( year, month, day, hour, minute, second)
		self.__noOfCyclicTasks = struct.unpack_from('<I', self._content, head+64)[0] # uint32
		self.__noOfSystemTasks = struct.unpack_from('<I', self._content, head+72)[0] # uint32		
		self.__profiledTargetType = str(self._content[head+84:head+99], encoding='ascii').rstrip('\x00')
		self.__maxPossibleEntries = int(self.__maxSize / (__class__.SIZEOF_RECORD_ENTRY + self.__additionalDatasizePerEntry))
		self.__systemTick = 1000 / (self.__tickFrequencyMhz)
		self.__offsetObjectEntry = head + 0xb6

		# now get the profiled objects:
		self.__objectTypes = dict()
		# 1. optional library functions
		self.__libraryFunctions = dict()
		offset = self.__offsetObjectEntry
		# read the dummy entries for library functions
		for n in range( len(libraryFunctionsList) ):
			ident, nameIndex = struct.unpack_from('<I20xI', self._content, offset)
			fname = libraryFunctionsList[nameIndex]
			self.__libraryFunctions.update( { ident : { 'name' : fname } } )
			self.__objectTypes.update( { -ident : BrProfilerObjectType.LIBRARY_FUNCTION } ) # use negative ident since lib.func. use their own counting
			offset += __class__.SIZEOF_LIBRARY_FUNCTION_ENTRY


		self.__objectTypes = dict()
		# 2. task classes
		self.__taskClasses = dict()
		for n in range(9):
			ident, cycleTime, toleranceTime, prio, name = struct.unpack_from('<I20xIII34s', self._content, offset)
			name = str(name, 'ascii').rstrip('\x00')
			objectPriority = 255- prio
			self.__taskClasses.update( { ident : { 'name' : name, 'cycleTime' : cycleTime, 'toleranceTime' : toleranceTime, 'objectPriority' : objectPriority } })
			self.__objectTypes.update( { ident : BrProfilerObjectType.TASK_CLASS } )
			offset += __class__.SIZEOF_TASKCLASS_ENTRY			

		# 3. cyclic tasks	
		self.__cyclicTasks = dict()
		for n in range(self.__noOfCyclicTasks):
			ident, taskclass, name = struct.unpack_from('<I20xI34s', self._content, offset)
			if ident:
				name = str(name, 'ascii').rstrip('\x00')
				self.__cyclicTasks.update( { ident : { 'name' : name, 'taskClass' : taskclass } } )
				self.__objectTypes.update( { ident : BrProfilerObjectType.CYCLIC_TASK } )
			offset += __class__.SIZEOF_OBJECT_ENTRY

		# 4. system tasks
		self.__systemTasks = dict()
		for n in range(self.__noOfSystemTasks):
			ident, prio, stackSize, freeStack, name = struct.unpack_from('<I20xIII34s', self._content, offset)
			if ident:
				name = str(name, 'ascii').rstrip('\x00')
				objectPriority = 255 - prio
				self.__systemTasks.update( { ident : { 'name' : name, 'objectPriority' : objectPriority, 'stackSize' : stackSize, 'freeStack' : freeStack} } )
				self.__objectTypes.update( { ident : BrProfilerObjectType.SYSTEM_TASK } )				
			offset += __class__.SIZEOF_SYSTEMOBJECT_ENTRY

		# 3. interrupt handler
		self.__interruptHandler = dict()
		while offset < self.__offsetRecords - __class__.MIN_SIZE_INTERRUPT_VECTOR:
			ident, interruptNr, length = struct.unpack_from('<I20xII', self._content, offset)
			offset += __class__.MIN_SIZE_INTERRUPT_VECTOR
			if ident != 0:
				name = bytearray()
				if length > 0:
					name = self._content[offset:offset+length]
					offset += length
				name = f'Interrupt {interruptNr} ({str( name, "ascii")})'
				self.__interruptHandler.update( { ident : { 'name' : name, 'interrupt' : interruptNr } } )
				self.__objectTypes.update( { ident : BrProfilerObjectType.INTERRUPT_VECTOR } )				
			else:
				break

		self.__entries = None


	@property
	def info(self) -> dict:
		return({
			"startDateOfProfiling" : self.__profilingTimestamp,
			"sg" : self.__sgType,
			"profiledTarget" : self.__profiledTargetType,
			"targetIdCode"   : self.__bridcode,
			"tickFrequencyMhz"  : self.__tickFrequencyMhz,
			"maxSize" : self.__maxSize,
			"additionalDataSizePerEntry" : self.__additionalDatasizePerEntry,
			"bufferForCreatedTask" : self.__bufferForCreatedTask,
			"bufferForCreatedUserTask" : self.__bufferForCreatedUserTask,
			"entries" : self.__maxPossibleEntries,
			"AR" : self.__arversion
		})


	@property
	def libraryFunctions(self) -> dict:
		return self.__libraryFunctions	
	

	@property
	def taskClasses(self) -> dict:
		return self.__taskClasses

	@property
	def cyclicTasks(self) -> dict:
		return self.__cyclicTasks

	@property
	def systemTasks(self) -> dict:
		return self.__systemTasks

	@property
	def interruptHandler(self) -> dict:
		return self.__interruptHandler


	@property
	def entries(self) -> list:
		if not(self.__entries):
			offset = self.__offsetRecords + self.__firstRecordAtOffset * __class__.SIZEOF_RECORD_ENTRY
			maxOffset = self.__offsetRecords + (self.__maxPossibleEntries-1) * __class__.SIZEOF_RECORD_ENTRY
			noOfValidEntries = self.__maxPossibleEntries
			if self.__lastRecordAtOffset > self.__firstRecordAtOffset:
				noOfValidEntries = self.__lastRecordAtOffset - self.__firstRecordAtOffset + 1
			else:
				noOfValidEntries = self.__lastRecordAtOffset + self.__maxPossibleEntries - self.__firstRecordAtOffset + 1
			self.__entries = list()
			startticks = None
			for n in range( noOfValidEntries):
				ticks, event, objectIdent = struct.unpack_from('<QII', self._content, offset)
				if not(startticks):
					startticks = ticks
				micros = (ticks-startticks) * self.__systemTick

				objectName = ''
				eventDescription = ''

				if event in BrLibraryFunctionEvents and objectIdent in self.__libraryFunctions:
					object = self.__libraryFunctions[objectIdent]
					objectName = object['name']
					eventDescription = BrLibraryFunctionEvents[event]

				else:
					objectType = self.__objectTypes.get(objectIdent, BrProfilerObjectType.UNKNOWN )
					if objectType == BrProfilerObjectType.UNKNOWN:
						objectName = 'Unknown Object'
					elif objectType == BrProfilerObjectType.TASK_CLASS:
						object = self.__taskClasses[objectIdent]
						objectName = object['name']
						eventDescription = BrTaskClassEvents.get( event, '' )
					elif objectType == BrProfilerObjectType.CYCLIC_TASK:
						object = self.__cyclicTasks[objectIdent]
						objectName = object['name']
						eventDescription = BrCyclicTaskEvents.get( event, '' )
					elif objectType == BrProfilerObjectType.SYSTEM_TASK:
						object = self.__systemTasks[objectIdent]
						objectName = object['name']
						eventDescription = BrSystemObjectEvents.get( event, '' )
					elif objectType == BrProfilerObjectType.INTERRUPT_VECTOR:
						object = self.__interruptHandler[objectIdent]
						objectName = object['name']
						eventDescription = BrInterruptHandlerEvents.get( event, '' ) 
					else:
						raise ValueError("Unknown profiler object type")


				self.__entries.append( BrProfilerEvent( micros, objectIdent, objectName, event, eventDescription) )
				if offset >= maxOffset:
					offset = self.__offsetRecords
				else:
					offset += __class__.SIZEOF_RECORD_ENTRY
		return self.__entries


	def __repr__(self) -> str:
		return f'BrProfilerDataFile'            
    


#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from ctypes import *
import datetime
import inspect
from .include import *
from .Object import PviObject
from .Error import PviError


# ----------------------------------------------------------------------------------
class Cpu(PviObject):

    def __init__( self, parent, name, **objectDescriptor ):
        if parent._type != T_POBJ_TYPE.POBJ_DEVICE and parent._type != T_POBJ_TYPE.POBJ_STATION:
            raise PviError(12009, self )                    
        super().__init__( parent, 'POBJ_CPU', name, **objectDescriptor) 
        self._downloaded = None
        self._progress = None  
  

    def __repr__(self):
        return f"Cpu( name={self._name}, linkID={self._linkID} )" 


    def _eventStatus( self, wParam, responseInfo ):
        """         
        handle status events
        """      
        self._result = PviReadResponse( wParam, None, 0 ) 
        if callable(self._statusChanged):
            self._statusChanged( responseInfo )


    def _eventDownloadStream( self, wParam, responseInfo : T_RESPONSE_INFO):    
        self._result = PviWriteResponse( wParam )
        if self._result == 0:
            if self._downloaded:
                sig = inspect.signature(self._downloaded)
                if len(sig.parameters) == 0:
                    self._downloaded()
                elif len(sig.parameters == 1):
                    self._downloaded(self)
        else:
            raise PviError(self._result, self)      


    def _eventProceeding( self, wParam, responseInfo : T_RESPONSE_INFO ):    
        proceedingInfo = T_PROCEEDING_INFO()
        self._result = PviReadResponse( wParam, byref(proceedingInfo), sizeof(proceedingInfo) )
        if self._result == 0:
            if self._progress:
                sig = inspect.signature(self._progress)
                if len(sig.parameters) == 1:
                    self._progress(int(proceedingInfo.Percent))
                elif len(sig.parameters) == 2:
                    self._progress(self, int(proceedingInfo.Percent))                    
        else:
            raise PviError(self._result, self)   


    def warmStart(self) -> None:
        '''
        Cpu.warmstart() -> None
        executes a warm restart
        '''
        s = create_string_buffer(b"ST=WarmStart")
        self._result = PviWrite( self._linkID, POBJ_ACC_STATUS, byref(s), sizeof(s), None, 0 )
        if self._result != 0:
            raise PviError(self._result, self)


    def coldStart(self) -> None:
        '''
        Cpu.coldStart() -> None
        executes a cold restart
        '''        
        s = create_string_buffer(b"ST=ColdStart")
        self._result = PviWrite( self._linkID, POBJ_ACC_STATUS, byref(s), sizeof(s), None, 0 )
        if self._result != 0:
            raise PviError(self._result, self)


    def downloadModule(self, data : bytes, **kwargs ):
        '''
        Cpu.downloadModule( data, **args )
        download given bytes as module

        Parameters:
        data : bytes  - data content
        kwargs : MT - module type e.g. 'BRT'
               MN - module name
               MV - module version
               LD - memory type e.g. 'Ram', 'Dram', 'Rom'
               downloaded - callback() or callback(Cpu) if module was downloaded
               progress - callback(p) or callback (Cpu, p) to show progress
        '''
        if not(isinstance(data, bytes)):
            raise TypeError(inspect.currentframe().f_code.co_name + " - data: only bytes accepted !")
        if 'MT' not in kwargs:
            kwargs.update( { 'MT' : 'BRT' } )
        if 'MN' not in kwargs:
            raise KeyError( inspect.currentframe().f_code.co_name + " - argument 'MN' is missing !")

        arguments = ''
        for key, value in kwargs.items():
            if key == 'downloaded':
                if callable(value):
                    self._downloaded = value
                else:
                    raise TypeError("only type function for argument 'downloaded' allowed !")
            elif key == 'progress':
                if callable(value):
                    self._progress = value
                else:
                    raise TypeError("only type function for argument 'progress' allowed !")
            else:
                arguments += f"{key}={value} "
        s = create_string_buffer(bytes(arguments,'ascii') + b'\0' + bytes(data) )            
        self._result = PviWriteRequest( self._linkID, POBJ_ACC_DOWNLOAD_STM, byref(s), sizeof(s), PVI_HMSG_NIL, SET_PVIFUNCTION, 0)
        if self._result:
            raise PviError( self._result, self )


    @property
    def modules(self):
        """     
        Cpu.modules : list of str
        get a list of all modules 
        """
        s = create_string_buffer(b'\000' * 4096)   
        self._result = PviRead( self._linkID, POBJ_ACC_LIST_MODULE, None, 0, byref(s), sizeof(s) )
        if self._result == 0:
            s = str(s, 'ascii').rstrip('\x00')
            return s.split('\t')
        else:
            raise PviError(self._result, self)


    @property
    def tasks(self):
        """     
        Cpu.tasks : list of str
        get a list of all tasks
        """
        s = create_string_buffer(b'\000' * 4096)   
        self._result = PviRead( self._linkID, POBJ_ACC_LIST_TASK, None, 0, byref(s), sizeof(s) )
        if self._result == 0:
            s = str(s, 'ascii').rstrip('\x00')
            return s.split('\t')
        else:
            raise PviError(self._result, self)


    @property
    def variables(self):
        """     
        Cpu.variables : list of str
        get a list of global variables
        """
        s = create_string_buffer(b'\000' * 65536)   
        self._result = PviRead( self._linkID, POBJ_ACC_LIST_PVAR, None, 0, byref(s), sizeof(s) )
        if( self._result == 0 ):
            s = str(s, 'ascii').rstrip('\x00')
            variables = [v.split(' ')[0] for v in s.split('\t')]
            return variables
        else:
            raise PviError(self._result, self)


    @property       
    def status(self) -> dict:
        """
        Cpu.status
        read the CPU status
        """
        st = super().status
        runState = { "WarmStart" : "RUN", "ColdStart" : "RUN", "Diagnose" : "DIAG", "Error" : "SERV", "Reset" : "SERV"}.get(st.get("ST"), "<unknown>")
        st.update({"RunState" : runState})
        return st
        

    @property
    def time(self) -> datetime.datetime:
        """
        Cpu: read the CPU's clock
        """
        t = struct_tm()       
        self._result = PviRead( self._linkID, POBJ_ACC_DATE_TIME , None, 0, byref(t), sizeof(t) )
        if self._result == 0:
            try:
                time =  datetime.datetime( year = t.tm_year+1900, month = t.tm_mon+1, day=t.tm_mday, hour=t.tm_hour, minute = t.tm_min, second = t.tm_sec)
            except ValueError:
                time = datetime.datetime( year = 1970, month = 1, day = 1 )            
            return time            
        else:
            raise PviError(self._result, self)
            


     

            



#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
from ctypes import *
import re
import struct
from collections import namedtuple, OrderedDict
from .Error import PviError
from .Object import PviObject
from .include import *


patternDataTypeInformation = re.compile(r"([A-Z]{2}=\w+)|(\{[^}]+\})")
patternStructureElementDefinition = re.compile(r"(\x2E[a-zA-z0-9_.]+)|([A-Z]{2}=\w+)")

StructMember = namedtuple( 'StructMember', ['name', 'vt', 'vn', 'vo', 'vl'])
'''
helper class for structure member 
'''


class VariableType():
    '''
    helper class for class Variable to parse data type information
    '''
    def __init__(self, object : PviObject ):
        '''
        object: Variable object
        '''        
        self._parentObject = object

        s = create_string_buffer(b'\000' * 64*1024) 
        object._result = PviRead( self._parentObject._linkID, POBJ_ACC_TYPE_INTERN, None, 0, s, sizeof(s) )
        if self._parentObject._result == 0:
            s = str(s, 'ascii').rstrip('\x00')
        else:
            raise PviError(self._parentObject._result, self._parentObject)

 
        self._members = None # let's assume a basic variable type
        '''
        list of all data type (structure) members if this variable is a struct
        '''

        self._parentObject._objectDescriptor.update({'VN' : 1, 'VL' : 1 })

        self._updateObjectDescriptor(s) # udpate objects descriptor
        self.vt = PvType( self._parentObject._objectDescriptor.get('VT')) 
        '''
        PVI Variable Type : PvType
        '''

        self.vn = int(self._parentObject._objectDescriptor.get('VN'))
        '''
        number of array elements : int
        '''

        self.vl = int(self._parentObject._objectDescriptor.get('VL'))
        '''
        variable byte length : int
        '''
         
    def _updateObjectDescriptor(self, s):
        '''
        update object descriptor and find all members if this PV is a structure
        '''
        self._innerStructOffsets = dict() # save vo of inner structs to correct their members' offsets
        for m in patternDataTypeInformation.finditer(s) : # update object descriptor
            ex = m.group()
            if ex.startswith('{'): # structure member definition
                if self._members == None:
                    self._members = list()
                self._updateMemberList(ex)
            else:
                self._parentObject._objectDescriptor.update({ex[0:2]: ex[3:]}) 
        del self._innerStructOffsets


    def _updateMemberList(self, s ):
        '''
        extract data type member's information
        s : member definition returned by POBJ_ACC_TYPE_INTERN
            e.g "{.ton.IN VT=boolean VL=1 VN=1 VO=16}"
        '''
        desc = dict() # member's object descriptor
        for m in patternStructureElementDefinition.finditer(s): # find matches
            ex = m.group()
            if ex.startswith('.'):
                desc.update({'name' : ex}) # member's name
            else:
                desc.update({ex[0:2]: ex[3:]}) # member's object descriptor

        vt = PvType(desc.get('VT'))
        vo = int(desc.get('VO'))
        vn = int(desc.get('VN'))
        vl = int(desc.get('VL'))        
        name = desc.get('name')
        parentStruct = name.rpartition('.')[0]

        if parentStruct in self._innerStructOffsets: # does member belong to an inner struct ?
            vo += self._innerStructOffsets[parentStruct]
 
        if vt == PvType.STRUCT: # is member itself an inner structure ?
            self._innerStructOffsets.update( { name :vo } )
        else:
            self._members.append(  StructMember( name, vt, vn, vo, vl ) )


    def _unpackRawData(self, data : bytes, vt : PvType, vl :int):
        '''
        unpacks data returned by PVI_ACC_DATA or PVI_EVENT_DATA to the appropiate Python data type
        data: ctype raw data        
        vt: PVI variable type
        vl: variable byte length 
        '''        
        if vt == PvType.BOOLEAN:
            return bool(struct.unpack('<?', data )[0])
        elif vt == PvType.U8:
            return struct.unpack('<B', data )[0] 
        elif vt == PvType.I8:
            return struct.unpack('<b', data)[0] 
        elif vt == PvType.U16:
            return struct.unpack('<H', data)[0] 
        elif vt == PvType.I16:
            return struct.unpack('<h', data)[0] 
        elif vt == PvType.U32 or vt == PvType.DT  or vt == PvType.DATE or vt == PvType.TOD:
            value = struct.unpack('<L', data)[0]
            if vt == PvType.DT:
                return datetime.datetime.fromtimestamp(value)
            elif vt == PvType.DATE:
                return datetime.date.fromtimestamp(value)                    
            elif vt == PvType.TOD:
                hour = int(value / 3600000)
                value %= 3600000
                minute = int(value / 60000)
                value %= 60000
                second = int(value/1000)
                millisecond = value % 1000
                return datetime.time( hour = hour, minute = minute, second = second, microsecond=millisecond*1000 ) 
            else:
                return value           
        elif vt == PvType.I32 or vt == PvType.TIME:
            value = struct.unpack('<l', data)[0]
            if vt == PvType.TIME:
                absValue = abs(value)
                sgn = 1 if value >= 0 else -1
                days = int(absValue / 86400000 )
                absValue %= 86400000
                seconds = int(absValue/1000)
                milliseconds = absValue % 1000
                return sgn * datetime.timedelta( days = days, seconds = seconds, microseconds=milliseconds*1000 )           
            else:
                return value
        elif vt == PvType.U64:
            return struct.unpack('<Q', data)[0] 
        elif vt == PvType.I64:
            return struct.unpack('<q', data)[0] 
        elif vt == PvType.F32:
            return struct.unpack('<f', data)[0] 
        elif vt == PvType.F64:
            return struct.unpack('<d', data)[0] 
        elif vt == PvType.STRING:
            return bytes(''.join( chr(_) for _ in data if _ != 0 ), 'ascii')
        elif vt == PvType.WSTRING:
            return ''.join( chr(int(data[_]) + int(data[_+1])) for _ in range(0,len(data),2) if int(data[_]))
        else:
            raise BaseException("not implemented")


    def readFromBuffer(self, buffer : bytes):
        '''
        gets the value from raw buffer data.
        returns a tuple in case of array.

        buffer : raw buffer data from POBJ_ACC_DATA
        '''
        arrayBuffers = ( buffer[_:_+self.vl] for _ in range(0,len(buffer), self.vl)) # split buffer into array elements
        result = list()

        for elementBuffer in arrayBuffers:
            if self.vt == PvType.STRUCT:
                structResult = OrderedDict()
                for m in self._members:
                    if m.vn == 1: # struct member is asingle value
                        buf = elementBuffer[m.vo : m.vo+m.vl]
                        value = self._unpackRawData( buf, m.vt, m.vl )
                    else: # struct member is an array
                        memberBuffers = ( elementBuffer[_:_+m.vl] for _ in range( m.vo, m.vo + m.vl*m.vn, m.vl)) # split buffer
                        value = tuple( self._unpackRawData( _, m.vt, m.vl ) for _ in memberBuffers )
                    structResult.update( { m.name : value } )
                result.append( structResult )
            else:
                result.append( self._unpackRawData( elementBuffer, self.vt, self.vl ) )

        if len(result) == 1: # single value
            return result[0]
        else:
            return tuple(result)


    def writeToBuffer(self, value):
        '''
        packs Python data type to byte buffer
        value: value
        '''     
        buffer = None

        if self.vn > 1 and not isinstance(value, list) and not isinstance(value, tuple): # variable is an array
            value = [value for _ in range(0, self.vn)] # create a list with identical values

        try:
            if self.vt == PvType.BOOLEAN:
                if self.vn == 1: # single value
                    buffer = c_bool(value)
                else: # array of BOOLEAN
                    buffer = (c_bool*self.vn)(*value)

            elif self.vt == PvType.U8:
                if self.vn == 1: # single value
                    buffer = c_uint8(value)
                else: # array of U8
                    buffer = (c_uint8*self.vn)(*value)

            elif self.vt == PvType.I8:
                if self.vn == 1: # single value
                    buffer = c_int8(value)
                else: # array of I8
                    buffer = (c_int8*self.vn)(*value)

            elif self.vt == PvType.U16:
                if self.vn == 1: # single value
                    buffer = c_uint16(value)
                else: # array of U16
                    buffer = (c_uint16*self.vn)(*value)

            elif self.vt == PvType.I16:
                if self.vn == 1: # single value
                    buffer = c_int16(value)
                else: # array of I16
                    buffer = (c_int16*self.vn)(*value)

            elif self.vt == PvType.U32:
                if self.vn == 1: # single value
                    buffer = c_uint32(value)
                else: # array of U32
                    buffer = (c_uint32*self.vn)(*value)

            elif self.vt == PvType.DT:
                if self.vn == 1: # single value
                    if type(value) == datetime.datetime:
                        buffer = c_uint32( int(value.timestamp()) )
                    elif type(value) == int:
                        buffer = c_uint32(value)
                else: # array of DT
                    if type(value[0]) == datetime.datetime:
                        v = (int(v.timestamp()) for v in value)
                        buffer = (c_uint32*self.vn)( *v )
                    elif type(value[0]) == int:
                        buffer = (c_uint32*self.vn)(*value)

            elif self.vt == PvType.DATE:
                if self.vn == 1: # single value
                    if type(value) == datetime.date:
                        value = datetime.datetime.combine(value, datetime.time())
                        buffer = c_uint32( int(value.timestamp()) )
                    elif type(value) == int:
                        buffer = c_uint32(value)
                else: # array of DATE
                    if type(value[0]) == datetime.date:
                        v = (int(datetime.datetime.combine(v, datetime.time()).timestamp()) for v in value)
                        buffer = (c_uint32*self.vn)( *v )
                    elif type(value[0]) == int:
                        buffer = (c_uint32*self.vn)(*value)            

            elif self.vt == PvType.TOD:
                if self.vn == 1: # single value
                    if type(value) == datetime.time:
                        buffer = c_uint32( value.hour * 3600000 + value.minute * 60000 + value.second * 1000 + int(value.microsecond / 1000) )
                    elif type(value) == int:
                        buffer = c_uint32(value)
                else: # array of TOD
                    if type(value[0]) == datetime.time:
                        v = (v.hour * 3600000 + v.minute * 60000 + v.second * 1000 + int(v.microsecond / 1000) for v in value)
                        buffer = (c_uint32*self.vn)( *v )
                    elif type(value[0]) == int:
                        buffer = (c_uint32*self.vn)(*value)

            elif self.vt == PvType.TIME:
                if self.vn == 1: # single value
                    if type(value) == datetime.timedelta: 
                        buffer = c_int32( int( value / datetime.timedelta(milliseconds=1)) )
                    elif type(value) == int:
                        buffer = c_int32(value)
                else: # array of TIME
                    if type(value[0]) == datetime.timedelta: 
                        v = (int( v / datetime.timedelta(milliseconds=1)) for v in value)
                        buffer = (c_int32*self.vn)( *v)
                    elif type(value[0]) == int:
                        buffer = (c_int32*self.vn)(value)

            elif self.vt == PvType.I32:
                if self.vn == 1: # single value
                    buffer = c_int32(value)
                else: # array of I32
                    buffer = (c_int32*self.vn)(*value)                

            elif self.vt == PvType.U64:
                if self.vn == 1: # single value
                    buffer = c_uint64(value)
                else: # array of U64
                    buffer = (c_uint64*self.vn)(*value)                

            elif self.vt == PvType.I64:
                if self.vn == 1: # single value 
                    buffer = c_int64(value)
                else: # array of I64
                    buffer = (c_int64*self.vn)(*value)                                

            elif self.vt == PvType.F32:
                if self.vn == 1: # single value
                    buffer = c_float(value)
                else: # array of F32
                    buffer = (c_float*self.vn)(*value)                

            elif self.vt == PvType.F64:
                if self.vn == 1: # single value
                    buffer = c_double(value)
                else: # array of F64
                    buffer = (c_double*self.vn)(*value)                

            elif self.vt == PvType.STRING:
                if self.vn == 1: # single value
                    if type(value) == bytes:
                        buffer = (c_char * self.vl)(*value)
                else: # array of STRING
                    if type(value[0]) == bytes:
                        buffer = ((c_char * self.vl) * self.vn)()
                        for n, v in enumerate(value):
                            s = (c_char*self.vl)(*v)
                            buffer[n] = s

            elif self.vt == PvType.WSTRING:
                ch = int(self.vl/2) # no of characters
                if self.vn == 1:
                    if type(value) == bytes:
                        value = value.decode()                
                    buffer = (c_wchar * ch )(*value)                
                else: # array of WSTRING
                    buffer = ((c_wchar * (ch)*self.vn))()
                    for n, v in enumerate(value):
                        if type(v) == bytes:
                            v = v.decode()
                        for j,c in enumerate(v):
                            buffer[n][j] = c

                    

            else:
                raise BaseException( "data type " + str(self.vt) + " not implemented")
            
        except IndexError:
            raise IndexError(f'wrong length writing {value}\nto {repr(self._parentObject)}')            
        except BaseException as e:
            raise e
                    
        if buffer == None:
            raise ValueError(f'wrong data type {type(value)} used\nwriting {repr(self._parentObject)}')

        return buffer


# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on ArSim 
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this example shows the usage of Pvi.py in tkinter
#

import tkinter as tk
import tkinter.ttk as ttk
from pvi import *


class ApplicationWindow(tk.Tk):
    """
    the application's main window
    """
    def __init__(self ):
        super().__init__()

        self.geometry('320x200')
        self.title('coffee machine') 
        self.pviLicense = 'unchecked';

        # setup widgets
        #self.columnconfigure( index = 0, weight= 1)
        self.rowconfigure(4, weight=1)

        self._lblTemperature = tk.Label(self, text="Temperature:")
        self._lblTemperature.grid( row= 0, column= 0 )              
        
        self._outputTemperature = tk.Label(self, text="00.0")
        self._outputTemperature.grid( row= 0, column= 1 )                       

        self._checkTemperatureOk = tk.IntVar(self, 0 )
        self._checkbtnTemperatureOk = ttk.Checkbutton( self, text= 'Ok', takefocus = False, variable = self._checkTemperatureOk )
        self._checkbtnTemperatureOk.grid( row=0, column=2)

        self._btnPowerOn = tk.Button( self, text ="Power On", command = self.actionPowerOn )   
        self._btnPowerOn.grid( row= 3, column= 0 )             

        self._btnPowerOff = tk.Button( self, text ="Power Off", command = self.actionPowerOff )   
        self._btnPowerOff.grid( row= 3, column= 1 )             

        
        # status bar
        self._lblStatusbar = tk.Label(self, text="on the way…", bd=1, relief=tk.SUNKEN, anchor=tk.W)
        self._lblStatusbar.grid( row= 4, column= 0,sticky='sew' )        

        self.connection = Connection()
        # all PVI objects must be registered hierarchically
        # line ANSL is the 'modern' way to access PLC variables
        # (compared to the older INA2000 line)
        #
        self.line = Line( self.connection.root, 'LNANSL', CD='LNANSL')
        self.device = Device( self.line, 'TCP', CD='/IF=TcpIp' )
        self.cpu = Cpu( self.device, 'myArsim', CD='/IP=127.0.0.1' )
        self.task1 = Task( self.cpu, 'mainlogic')

        self.gHeating = Variable( self.task1, 'gHeating', RF=200 )
        self.gHeating.valueChanged = self.callbackHeating
       
        # we then register a variable to switch on the machine
        self.switchOnOff = Variable( self.task1, 'gMainLogic.cmd.switchOnOff' )        

        self.after( 10, self.update) 
        self.after( 1000, self.checkPviLicense )                      

    def actionPowerOn(self):
        self.switchOnOff.value = True

    def actionPowerOff(self):
        self.switchOnOff.value = False
           
    def callbackHeating(self, value):
        self._outputTemperature.config( text = round( float(value.get('.status.actTemp')), 1) )
        self._checkTemperatureOk.set( int(value.get('.status.setTempOK')) )
        
    def update(self):
        try:
            self.connection.doEvents() # execute PVI event loop
        except:
            pass
        self.after( 100, self.update)  

        
    def checkPviLicense(self):
        li = self.connection.license[0]
        self._lblStatusbar.config( text = 'PVI license: ' + li )        
        self.after( 1000 if li == 'undefined' else 30000 , self.checkPviLicense )                


app = ApplicationWindow()
app.mainloop()


#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from ctypes import *

# C/C++ struct tm
# struct tm
# {
#     int tm_sec; // seconds after the minute - [0,59]
#     int tm_min; // minutes after the hour - [0,59]
#     int tm_hour; // hours since midnight - [0,23]
#     int tm_mday; // day of the month - [1,31]
#     int tm_mon; // months since January - [0,11]
#     int tm_year; // years since 1900
#     int tm_wday; // days since Sunday - [0,6]
#     int tm_yday; // days since January 1 - [0,365]
#     int tm_isdst; // daylight savings time flag
# };


class struct_tm(Structure):
    _fields_ = [
        ("tm_sec", c_int32),
        ("tm_min", c_int32),
        ("tm_hour", c_int32),
        ("tm_mday", c_int32),
        ("tm_mon", c_int32),
        ("tm_year", c_int32),
        ("tm_wday", c_int32),
        ("tm_yday", c_int32),                                                
        ("tm_isdst", c_int32)
    ]
 


# simple1.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on ArSim 
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this simple example just registers a variable, reads its value and then
# exit after a few seconds
#


from time import sleep
import datetime
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# line ANSL is the 'modern' way to access PLC variables
# (compared to the older INA2000 line)
#
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )
task1 = Task( cpu, 'mainlogic')
# we register a variable with refresh time of 0 seconds by default
temperature = Variable( task1, 'gHeating.status.actTemp' )
# we register a callback to get variable's value
temperature.valueChanged = lambda value : print(f'{temperature.name} = {value}')

startTime = datetime.datetime.now()
run = True

def cpuErrorChanged( error : int ):
    global run

    if error != 0:
        raise PviError(error)

cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    if datetime.datetime.now() - startTime > datetime.timedelta(seconds = 10):
        run = False # exit the loop
    else:
        sleep(0.1)








# simple3.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on PP65 with AR 3.x
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this simple example just registers a variable for reading and another
# for writing. In fact we switch on the 'coffee machine' and watch its temperature ...
#
# This is similar to simple2.py but since we use a control running < AR 4.x
# ANSL is not available here and we change to good old INA2000


from time import sleep
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# ANSL is not available and we use INA2000
#
line = Line( pviConnection.root, 'LNINA', CD='LNINA2')
device = Device( line, 'TCP', CD='/IF=TcpIp /SA=113' ) # always use a unique node number for PVI client even though you won't use node numbers
cpu = Cpu( device, 'myPP65', CD='/DAIP=10.49.40.222' )

# alternative: use 'INA node number' instead of IP address. INA node numbers must be unique in network !
# cpu = Cpu( device, 'myPP65', CD='/DA=32' )

task1 = Task( cpu, 'mainlogic')
# we register the variable containing the temperature 
temperature = Variable( task1, 'gHeating.status.actTemp', RF=2000 )
# we register a callback to get temperature value
temperature.valueChanged = lambda value : print(f'\rTemperature = {value}', end="")
# we then register a variable to switch on the machine
switch = Variable( task1, 'gMainLogic.cmd.switchOnOff' )

run = True
warmUp = False
coolDown = False


def cpuErrorChanged( error : int ):
    global run

    if error != 0:
        raise PviError(error)

cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called

    if temperature.readable and switch.writable:
        if temperature.value < 25 and not warmUp and not coolDown:
            switch.value = 1 # switch on machine
            warmUp = True
            coolDown = False
            print('warming up...')
        elif temperature.value > 70 and warmUp:
            warmUp = False
            coolDown = True
            switch.value = False # switch off
            print('\ncooling down...')        
        if coolDown and not warmUp and temperature.value < 25:
            print("\nit's cool guys !")
            run = False # exit the loop
        else:
            sleep(0.1)








# set_ip_address.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# this example searches for a specific B&R plc in local network with PVI's 'SNMP' line
# and changes its IP address
#

from time import sleep
import json
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
#

line = Line( pviConnection.root, 'LNSNMP', CD='LNSNMP')
device = Device( line, 'Device', CD='/IF=snmp /RT=1000' )

def deviceErrorChanged( error : int ):
    global run

    if error == 0:
        macs = [ x['name'] for x in device.externalObjects if x['type'] == 'Station']
        for mac in macs:
            if mac == '00-60-65-14-d6-da':
                station = Station( device, 'station', CD=f'/CN={mac}' )
                # there is no freedom how to name the internal SNMP variables
                ipAddress = Variable( station, 'ipAddress')
                ipAddress.value = b'192.168.0.10' # must be a byte string
                ipAddress.kill()
                run = False
    device.kill()
    line.kill()
    run = False

device.errorChanged = deviceErrorChanged

run = True

while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)




# list_objects1.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on ArSim 
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this example lists objects with 'global scope' (modules, task and global variables)
# from 'coffe machine' cpu and returns status information about them
#


from time import sleep
from pprint import pprint
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# line ANSL is the 'modern' way to access PLC variables
# (compared to the older INA2000 line)
#
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )


run = True

def cpuErrorChanged( error : int):
    global run

    if error == 0:
        # read content
        allObjects = cpu.externalObjects

        # read module names
        moduleNames = [_['name'] for _ in allObjects if _['type'] == 'Module' ]  

        modules = list()
        for name in moduleNames:  # read the modules' status
            module = Module( cpu, name )
            status = { "Name" : name }
            status.update( module.status )
            modules.append(status)
            module.kill()

        # read task names
        taskNames = [ _['name'] for _ in allObjects if _['type'] == 'Task'] 
        tasks = list()
        for name in taskNames:  # read the tasks' status
            task = Task( cpu, name )
            status = { "Name" : name }
            status.update( task.status )
            tasks.append(status)
            task.kill()

        # read names of global variables
        globalVarnames = [ _['name'] for _ in allObjects if _['type'] == 'Pvar'] 
        globalVariables = list()
        for name in globalVarnames:  # read the variables' status, data type and value
            variable = Variable( cpu, name )
            status = { "Name" : name }
            status.update( variable.status )
            status.update( { "Value" : variable.value} )
            status.update( { "Type" : variable.dataType} )            
            globalVariables.append(status)
            variable.kill()

 #      write content to file
        with open('content.txt', 'w') as f:
            f.write("Modules =\n")
            pprint(modules, stream=f, indent = 4)
            f.write("\nTasks =\n")
            pprint(tasks, stream=f, indent = 4)            
            f.write("\nglobalVars =\n")
            pprint(globalVariables, stream=f, indent = 4)  
        print('content.txt was created !')                      
        run = False
    
    elif error:
        raise PviError(error)


cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)








#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from ctypes import *
from .include import *
from .Object import PviObject
from .Error import PviError


class Task(PviObject):
    def __init__( self, parent, name, **objectDescriptor ):
        if parent._type != T_POBJ_TYPE.POBJ_CPU:
            raise PviError(12009, self)
        objectDescriptor.update({'CD':name})
        super().__init__( parent, 'POBJ_TASK', name, **objectDescriptor)

    def start(self):
        '''
        Task: start a task
        '''
        s = create_string_buffer(b"ST=Start")
        self._result = PviWrite( self._linkID, POBJ_ACC_STATUS, byref(s), sizeof(s), None, 0 )
        if self._result != 0:
            raise PviError(self._result, self)  

    def resume(self):
        '''
        Task: resume a stopped task
        '''
        s = create_string_buffer(b"ST=Resume")
        self._result = PviWrite( self._linkID, POBJ_ACC_STATUS, byref(s), sizeof(s), None, 0 )
        if self._result != 0:
            raise PviError(self._result, self)                         

    def stop(self):
        '''
        Task: stop task
        '''
        s = create_string_buffer(b"ST=Stop")
        self._result = PviWrite( self._linkID, POBJ_ACC_STATUS, byref(s), sizeof(s), None, 0 )
        if self._result != 0:
            raise PviError(self._result, self)                  

    @property
    def variables(self):
        """     
        Task.variables : list of str
        get a list of local variables
        """
        s = create_string_buffer(b'\000' * 65536)   
        self._result = PviRead( self._linkID, POBJ_ACC_LIST_PVAR, None, 0, byref(s), sizeof(s) )
        if( self._result == 0 ):
            s = str(s, 'ascii').rstrip('\x00')
            variables = [v.split(' ')[0] for v in s.split('\t')]
            return variables
        else:
            raise PviError(self._result, self)


    def __repr__(self):
        return f"Task( name={self._name}, linkID={self._linkID} )"




# modules.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# in this exammple we search for BR files (*.br) in a folder and read the type their content
#

import os
from pvi.utils import BrFile

# 'Temp' folder contains 'anonymous' file names and makes it difficult to determine the content
path_of_the_directory = r'C:\projects\CoffeeMachine\Temp\Simulation\Simulation\ARsim\RPSHD\USERROM'

ext = ('.br')
filenames = [f for f in os.listdir(path_of_the_directory) if f.endswith(ext)]

for filename in filenames:
    module = BrFile(path_of_the_directory + '\\' + filename)
    print( f'content of {filename} is {module.fileType}' )



# test against 'coffee machine' in AS 4.1.17.113 SP
from pathlib import Path
import time
import sys
import hashlib

pviPath = str(Path(__file__).parents[1])
cwd = str(Path(__file__).parents[0])

sys.path.append( pviPath )

from pvi import *

pviConnection = Connection() # start a Pvi connection
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )

run = True

def cpuErrorChanged( error : int):
    global run

     
    if error == 0:
        # read content
        allObjects = cpu.externalObjects

        dataType = None
        tasks = dict()
        for taskName in ['mainlogic', 'feeder', 'conveyor', 'brewing', 'heating', 'visCtrl','visAlarm', 'visTrend' ] :
            task = Task( cpu, taskName )
            variableNames = [ _['name'] for _ in task.externalObjects]
            variables = list()                    
            for variableName in variableNames:
                variable = Variable(task, variableName)

                try:
                    dataType = variable.dataType
                    value = variable.value
                except PviError as e:
                    print(e)
                variables.append( { "Name" : variableName, "Type" : dataType} )
                variable.kill()
            tasks.update({ taskName : variables })
            task.kill

        globalVariableNames = [ _['name'] for _ in allObjects if _['type'] == 'Pvar'] # read names of global variables
        variables = list()
        for variableName in globalVariableNames:
            variable = Variable(cpu, variableName)
            try:
                dataType = variable.dataType
                value = variable.value
            except PviError as e:
                print(e)
            variables.append( { "Name" : variableName, "Type" : dataType } )
            variable.kill()
        
        result = str(tasks) + str(variables)
        h = hashlib.sha256( result.encode() )
        
        if h.hexdigest() == '0ecb1036ba0bc689dc64de86ee1621fc7499da508d038129f258bfadaddf5293':
            print("pass !")
        else:
            print("failed !")

        run = False
    
    elif error:
        raise PviError(error)


cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    time.sleep(0.1)






# profiler1.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# extract all profiler data from a *.pd file and save it to csv
# a separate *.txt file is created with object information
#

import csv
import locale
from pprint import PrettyPrinter
from pvi.utils import BrProfilerDataFile

folder = r'C:\Temp\BuR_SDM_Sysdump_2023-07-25_10-35-25.tar\BuR_SDM_Sysdump_2023-07-25_10-35-25\Data Files'
file = r'BuR_SDM_profiling_2023-07-25_09-04-41.pd'

profilerData = BrProfilerDataFile(folder + '\\' + file)

with open(f'{file}.txt', 'w', newline = '') as f: 
    pp = PrettyPrinter( indent = 4, stream = f )
    print( file,  '\n\nSummary:\n', file = f)
    pp.pprint( profilerData.info )
    print( '\n\ntask classes:\n', file = f )
    pp.pprint( profilerData.taskClasses )                
    print( '\n\ncyclic tasks:\n', file = f )
    pp.pprint( profilerData.cyclicTasks )                
    print( '\n\nsystem tasks:\n', file = f )
    pp.pprint( profilerData.systemTasks )                
    print( '\n\ninterrupt handler:\n', file = f )
    pp.pprint( profilerData.interruptHandler )                
    print( '\n\nprofiled library functions:\n', file = f )
    pp.pprint( profilerData.libraryFunctions )                


    
with open(f'{file}.csv', 'w', newline = '') as f: 
    writer = csv.writer(f, delimiter= ';', quotechar="'" )
    writer.writerow ( ["Nr", "time", "object name", "ident","event", "event description"])            
    for n, e in enumerate(profilerData.entries):
        writer.writerow( ( n+1, locale.format_string( '%.6f', e.microseconds), e.objectName, hex(e.objectIdent), hex(e.event), e.eventDescription ) ) 




#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from .BrFile import *

class BrDataObjectFile(BrFile):
    '''
    class for a *.br file containing data object
    '''
    def __init__(self, filename: str):
        super().__init__(filename)
        if self._fileType != BrFileType.DATA_OBJECT:
            raise TypeError(f'content is not a B&R data module (Type is {self._fileType})')

    @property
    def version(self) -> int:
        v = struct.unpack_from('<B', self._content, 0x82) + 256 * struct.unpack_from('<B', self._content, 0x82)
        return v

    @property
    def xmlHeader(self):
        header = bytearray()
        for x in self._content[0x30:]:
            if x != 0:
                header.append(x)
            else:
                break
        return header.decode('ascii')        


# list_objects3.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on ArSim 
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this example lists all local variables of a specific task and their content
#


from time import sleep
from pprint import pprint
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# line ANSL is the 'modern' way to access PLC variables
# (compared to the older INA2000 line)
#
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )
task = Task( cpu, 'conveyor' )


run = True

def cpuErrorChanged( error : int):
    global run

    if error == 0:
        variableNames = [ _['name'] for _ in task.externalObjects]
        variables = list()                        
        for variableName in variableNames:
            variable = Variable(task, variableName)
            try:
                dataType = variable.dataType
                value = variable.value
            except PviError as e:
                print(e)
            variables.append( { "Name" : variableName, "Type" : dataType, "Value" : value} )
            variable.kill()

        # write content to file
        with open('content.txt', 'w') as f: 
            pprint(variables, stream=f, indent = 4)
        print('content.txt was created !')                      
        run = False
    
    elif error:
        raise PviError(error)


cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)








# linknode1.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is X20CP3584 with a X20DO6322 running AR G4.09
#
# this simple example just toggles an forced output
# remark: force status must be set with Automation Studio !
#


from time import sleep
import datetime
from pvi import *


class Output( Variable ):
    '''
    convinience class to make code more readable
    '''
    def __init__(self, cpu, linkNode ):
        super().__init__( cpu, linkNode, CD=f'"{linkNode}"' )
        self._state = False

    @property
    def forceable(self) -> bool:
        return True if self.status['FC'] == '1' else False
        
    def set(self):
        self._state = True
        self.value = self._state

    def reset(self):
        self._state = False        
        self.value = self._state

    def toggle(self) -> bool:
        if self._state:
            self.reset()
        else:
            self.set()
        return self._state
 
        
        
pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myCpu', CD='/IP=10.49.40.221' )
# syntax for linknode variables: GUID 3d1b8b36-111a-4f29-ac05-a6f0e6c76a80
output1 = Output( cpu, 'F+%QX.IF6.ST1.DigitalOutput01')


run = True
cpuInRun = False

def cpuErrorChanged( error : int ):
    global run
    global cpuInRun

    if error != 0:
        raise PviError(error)
    cpuInRun = cpu.status['RunState'] == 'RUN'
   

    
cpu.errorChanged = cpuErrorChanged
startTime = datetime.datetime.now()

while run:
    if output1.writable and cpuInRun:
        if datetime.datetime.now() - startTime > datetime.timedelta(seconds=5):
            startTime = datetime.datetime.now()         
            if output1.forceable:
                print( output1.toggle() )
            else:
                print('output cannot be forced !')

    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)








#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from ctypes import *
from .include import *
from .Error import PviError
from .Object import PviObject
import datetime

class Connection():
    '''
    class representing a connection to PVI manager
    '''
    # ----------------------------------------------------------------------------------
    def __init__(self, debug = False, timeout = 5 ):
        self._startTime = datetime.datetime.now()
        self._pviTrialTimeCheck = datetime.timedelta(seconds=timeout +1)
        self._debug = debug
        self._objectsArranged = False
        self._pviObjects = []
        self._rootObject = PviObject(None, 'POBJ_PVI', '@Pvi')
        self._rootObject._pviConnection = self
        self.link(self._rootObject)
        self._linkIDs = {}
        self._result = PviInitialize( timeout, 0, "", None )
        if self._result == 0:
            # set global events
            for m in (POBJ_EVENT_PVI_CONNECT, POBJ_EVENT_PVI_DISCONN, POBJ_EVENT_PVI_ARRANGE):
                self._result = PviSetGlobEventMsg( m, PVI_HMSG_NIL, SET_PVIFUNCTION, 0 )
                if self._result != 0:
                    raise PviError(self._result)
        else:
            raise PviError(self._result)

    # ----------------------------------------------------------------------------------
    def __repr__(self):
        return f"Connection()"

    # ----------------------------------------------------------------------------------
    def __del__(self):
        if self._debug:
            print("PviDeinitialize")
        PviDeinitialize()

    # ----------------------------------------------------------------------------------
    @property
    def root(self):
        '''
        return base object 'Pvi'
        '''
        return self._rootObject

    @property
    def license(self):
        """
        read license information
        """
        li = T_PVI_INFO_LICENCE()
        self._result = PviRead( self.root._linkID, POBJ_ACC_INFO_LICENCE  , None, 0, byref(li), sizeof(li) )
        if self._result == 0:
            try: 
                state = ('undefined', 'trial', 'runtime', 'developer', 'locked')[li.PviWorkState0]
                hardware = 'B&R IPC' if li.PviWorkState1 and 0x01 else 'PC'
                burlicence = 'B&R License' if li.PviWorkState1 and 0x02 else ''
                dongle = 'Pvi Dongle' if li.PviWorkState1 and 0x04 else ''
                return state, hardware, burlicence, dongle, str(li.LcName)
            except IndexError:
                pass
        else:
            raise PviError(self._result, self )
        return ('undefined', '', '', '', '', '' )

    # ----------------------------------------------------------------------------------
    def link(self, *args ):
        """
        registers new PviObject(s)

        *args : PviObject of list of PviObjects
        """
        for o in args:
            if isinstance(o, PviObject):
                self._pviObjects.append(o)
                if self._objectsArranged:
                    o._createAndLink(self)
            elif isinstance(o, list ):
                for oo in o:
                    self.link(oo)
            else:
                raise ValueError("only PviObject or list of PviObjects allowed !")
    # ----------------------------------------------------------------------------------

    def findObjectByName(self, name : str) ->PviObject :
        for o in self._pviObjects:
            if o.name == name:
                return o
        return None

    # ----------------------------------------------------------------------------------

    def findObjectByLinkID(self, linkID : DWORD) ->PviObject :
        for o in self._pviObjects:
            if o._linkID == linkID:
                return o
        return None        

    # ----------------------------------------------------------------------------------
    def _eventPviConnect( self, wParam, responseInfo ):
        """
        handle PVI connect event
        """
        if self._debug:
            print("POBJ_EVENT_PVI_CONNECT")
        self._result = PviReadResponse( wParam, None, 0 )     

    # ----------------------------------------------------------------------------------
    def _eventPviDisconnect( self, wParam, responseInfo ):
        """
        handle PVI disconnect event
        """  
        self._result = PviReadResponse( wParam, None, 0 )
        self._pviObjects.clear()
        self._linkIDs.clear()
        self._objectsArranged = False

    # ----------------------------------------------------------------------------------
    def _eventPviArrange( self, wParam, responseInfo ): 
        """
        handle pvi arrange event
        """
        self._pviTrialTimeCheck = None

        self._result = PviReadResponse( wParam, None, 0 )                
        if self._debug:
            print("POBJ_EVENT_PVI_ARRANGE")

        # create and link objects
        for po in self._pviObjects:
            po._createAndLink(self)
        self._objectsArranged = True

        if self._debug:
            license = self.license            
            print(f'PVI license: {license[0]} - running on {license[1]}' ) 

    # ----------------------------------------------------------------------------------
    def _eventOther( self, wParam, responseInfo ):
        """
        handle other events
        """
        self._result = PviReadResponse( wParam, None, 0 )

    # ----------------------------------------------------------------------------------
    def doEvents(self):
        """         
        event loop - must be cyclically called
        """
        wParam = WPARAM()
        lParam = LPARAM()
        hMsg = HANDLE()
        dataLen = c_uint32()        
        self._result = PviGetNextResponse( byref(wParam), byref(lParam), byref(hMsg), None )

        if wParam.value != 0:
            responseInfo = T_RESPONSE_INFO()
            self._result = PviGetResponseInfo( wParam, None, byref(dataLen), byref(responseInfo), sizeof(responseInfo) )

            if self._debug:
                if responseInfo.ErrCode != 0:
                    po = self.findObjectByLinkID(responseInfo.LinkID)
                    print(f'{po.name} : nMode = {responseInfo.nMode}, nType = {responseInfo.nType}, ErrCode = {responseInfo.ErrCode}')

            if responseInfo.nType == POBJ_EVENT_PVI_CONNECT:
                self._eventPviConnect( wParam, responseInfo )
            elif responseInfo.nType == POBJ_EVENT_PVI_DISCONN:
                self._eventPviDisconnect( wParam, responseInfo )
            elif responseInfo.nType == POBJ_EVENT_PVI_ARRANGE:
                self._eventPviArrange( wParam, responseInfo )
            elif responseInfo.nType == POBJ_ACC_TYPE:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventDataType( wParam, responseInfo )
                else:
                    raise ValueError("linkID not found !")
            elif responseInfo.nType == POBJ_EVENT_DATA:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventData( wParam, responseInfo )
                else:
                    raise ValueError("linkID not found !")                    
            elif responseInfo.nType == POBJ_ACC_UPLOAD_STM:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventUploadStream( wParam, responseInfo, dataLen.value ) 
                else:
                    raise ValueError("linkID not found !")
            elif responseInfo.nType == POBJ_ACC_DOWNLOAD_STM:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventDownloadStream( wParam, responseInfo ) 
                else:
                    raise ValueError("linkID not found !") 
            elif responseInfo.nType == POBJ_EVENT_STATUS or responseInfo.nType == POBJ_ACC_STATUS:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventStatus( wParam, responseInfo )
                else:
                    raise ValueError("linkID not found !")                   
            elif responseInfo.nType == POBJ_EVENT_PROCEEDING:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventProceeding( wParam, responseInfo ) 
                else:
                    raise ValueError("linkID not found !")   
            elif responseInfo.nType == POBJ_ACC_LN_XML_LOGM_DATA:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventUploadLogData( wParam, responseInfo, dataLen.value ) 
                else:
                    raise ValueError("linkID not found !") 
            elif responseInfo.nType == POBJ_ACC_MOD_DATA:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    po._eventUploadModData( wParam, responseInfo, dataLen.value ) 
                else:
                    raise ValueError("linkID not found !")                                                                    
            elif responseInfo.nType == POBJ_EVENT_ERROR:
                po = self._linkIDs.get(responseInfo.LinkID, None )
                if po:
                    if self._debug and responseInfo.ErrCode != 0:
                        print( f'{repr(po)}: error {responseInfo.ErrCode}')
                    po._eventError( wParam, responseInfo )
                else:
                    raise ValueError("linkID not found !")
            else:
                self._eventOther( wParam, responseInfo )
                if self._debug:
                    po = self.findObjectByLinkID(responseInfo.LinkID)
                    print(f'{po.name} : nMode = {responseInfo.nMode}, nType = {responseInfo.nType}, ErrCode = {responseInfo.ErrCode}')                
        
        if self._pviTrialTimeCheck:
            if datetime.datetime.now() - self._startTime > self._pviTrialTimeCheck:
                raise PviError(12040)





# list_objects2.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on ArSim 
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this example lists global and local variables and their content
#


from time import sleep
from pprint import pprint
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# line ANSL is the 'modern' way to access PLC variables
# (compared to the older INA2000 line)
#
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )


run = True

def cpuErrorChanged( error : int):
    global run

     
    if error == 0:
        # read content
        allObjects = cpu.externalObjects
        taskNames = [ _['name'] for _ in allObjects if _['type'] == 'Task'] # read task names

        tasks = dict()
        for taskName in taskNames:
            task = Task( cpu, taskName )
            variableNames = [ _['name'] for _ in task.externalObjects]
            variables = list()                    
            for variableName in variableNames:
                variable = Variable(task, variableName)
                try:
                    dataType = variable.dataType
                    value = variable.value
                except PviError as e:
                    print(e)
                variables.append( { "Name" : variableName, "Type" : dataType, "Value" : value} )
                variable.kill()
            tasks.update({ taskName : variables })
            task.kill
        # write content to file
        with open('content.txt', 'w') as f: 
            pprint(tasks, stream=f, indent = 4)

        globalVariableNames = [ _['name'] for _ in allObjects if _['type'] == 'Pvar'] # read names of global variables
        variables = list()
        for variableName in globalVariableNames:
            variable = Variable(cpu, variableName)
            try:
                dataType = variable.dataType
                value = variable.value
            except PviError as e:
                print(e)
            variables.append( { "Name" : variableName, "Type" : dataType, "Value" : value} )
            variable.kill()


        # append the content to existing file
        with open('content.txt', 'a') as f: 
            f.write("\nglobalVars =\n")
            pprint( variables, stream=f) 

        print('content.txt was created !')                      
        run = False
    
    elif error:
        raise PviError(error)


cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)








#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from ctypes import *
import re
import inspect
from .include import *
from .Error import PviError

class PviObject():
    __patternParameterPairs = re.compile(r"\s*([A-Z]{2}=\w*)\s*")

    """
    the base of all objects
    """
    def __init__(self, parent , type, name, **objectDescriptor ):
        '''
        > parent: parent PviObject
        > type:    POBJ_PVI, POBJ_LINE, POBJ_DEVICE, POBJ_STATION, POBJ_CPU, POBJ_MODULE, POBJ_TASK, POBJ_PVAR
        > name: name of object in PVI's hierarchy
        > objectDescriptor: CD=""
        '''
        parentName = re.findall('(\S+)',parent.name)[0]+'/' if parent else ''
        self._name = f'{parentName}{name}'
        self._linkID = 0
        self._objectDescriptor = objectDescriptor
        self._type = T_POBJ_TYPE[type]
        self._result = int(0)      
        self._pviError = int(0)
        self._errorChanged = None
        self._debug = False
        self._parent = parent
        if parent: # all objects but '@Pvi' have a parent
            self._pviConnection = parent._pviConnection
            self._pviConnection.link(self)
            self._debug = self._pviConnection._debug
            
    def __hash__(self):
        return hash( (self._name, self._type) )

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __str__(self):
        return f"{self._name}"

    def __repr__(self):
        return f"PviObject( name={self._name}, linkID={self._linkID} )"

    @property
    def name(self) -> str:
        '''
        PviObject: hierarchical object name
        '''
        return self._name

    @property
    def objectName(self) -> str:
        '''
        PviObject: object name
        '''            
        x = self._name.rpartition('/')
        try:
            return x[2]
        except IndexError:
            return self._name
        

    @property
    def descriptor(self) -> str:
        '''
        PviObject: object descriptor string
        '''
        return self._objectDescriptor

    @property
    def errorChanged(self) -> callable:
       '''
       PviObject: set callback for 'error changed'
       > cb: callback( PviObject, int ) or callback( int )
       '''
       return self._errorChanged

    @errorChanged.setter
    def errorChanged(self, cb : callable):
        '''
        PviObject: set callback for 'error changed'
        > cb: callback( PviObject, int ) or callback( int )
        '''
        if callable(cb):
            self._errorChanged = cb
        else:
            raise TypeError("only callable allowed for 'errorChanged'")


    def _eventData( self, wParam, responseInfo ):
        """
        handle data events
        """
        self._result = PviReadResponse( wParam, None, 0 )
        if callable(self._errorChanged):
            self._errorChanged(0)             

    def _eventDataType( self, wParam, responseInfo ):
        """
        handle data type events
        """        
        self._result = PviReadResponse( wParam, None, 0 ) 
       

    def _eventUploadStream( self, wParam, responseInfo, dataLen : int ):      
        self._result = PviReadResponse( wParam, None, 0 ) 


    def _eventStatus( self, wParam, responseInfo ):
        pass


    def _eventError( self, wParam, responseInfo ):
        """         
        handle error events
        """      
        self._result = PviReadResponse( wParam, None, 0 )
        if callable(self._errorChanged):
            sig = inspect.signature(self._errorChanged)
            if len(sig.parameters) == 1:
                self._errorChanged(responseInfo.ErrCode)
            elif len(sig.parameters) == 2:
                self._errorChanged( self, responseInfo.ErrCode)
 
    def _createAndLink(self, pvi):
        descriptor_items = []
        for key, value in self._objectDescriptor.items():
            quote = '"' if re.search( "[\/\.\s]", str(value) ) is not None else ''
            descriptor_items += [f'{key}={quote}{value}{quote}']
        descr = ' '.join(descriptor_items) 
        linkID = DWORD(0)
        linkDescriptor = None
        if self._type == T_POBJ_TYPE.POBJ_CPU or self._type == T_POBJ_TYPE.POBJ_MODULE:
            linkDescriptor = b'EV=ep' # need this for downloading proceeding info
        self._result = PviCreate( byref(linkID), bytes(self._name, 'ascii'),
            self._type, bytes(descr, 'ascii'), PVI_HMSG_NIL, SET_PVIFUNCTION, 0, linkDescriptor)
        if self._result == 0: # object creation successful
            self._linkID = linkID.value
            # if self._type == T_POBJ_TYPE.POBJ_PVAR: # read variable's data type
            #     PviReadRequest( self._linkID, POBJ_ACC_TYPE, PVI_HMSG_NIL, SET_PVIFUNCTION, 0 )
            pvi._linkIDs[self._linkID] = self # store object for backward reference  
        else:
            print( f"PviCreate {self.name} = {self._result}")
            raise PviError(self._result, self)
        return self._result                


    @property
    def externalObjects(self):
        """     
        PviObject.externalObjects : list of dict
        get a list of external objects
        # only available with ANSL, not with INA2000
        """    
        s = create_string_buffer(b'\000' * 65536)   
        self._result = PviRead( self._linkID, POBJ_ACC_LIST_EXTERN, None, 0, byref(s), sizeof(s) )
        if self._result == 0:
            s = str(s, 'ascii').rstrip('\x00')
            li1 = [r.split(' OT=') for r in s.split('\t')]
            li2 = [{ 'name': r[0], 'type' : r[1] } for r in li1]
            return li2
        else: 
            raise PviError(self._result, self)


    @property       
    def status(self) -> dict:
        """
        PviObject.status
        read the object's status () 
        """
        s = create_string_buffer(b'\000' * 64)             
        self._result = PviRead( self._linkID, POBJ_ACC_STATUS, None, 0, byref(s), sizeof(s) )
        st = dict()        
        if self._result == 0:
            s = str(s, 'ascii').rstrip('\x00')
            matches = PviObject.__patternParameterPairs.findall(s)
            if matches:
                for m in matches: 
                    token = m.split("=")
                    st.update( {token[0]:token[1]})
        else:
            raise PviError(self._result, self)  
        return st    

    
    @status.setter
    def status(self, st : bytes):
        s = create_string_buffer(st)
        self._result = PviWrite( self._linkID, POBJ_ACC_STATUS, byref(s), sizeof(s), None, 0 ) 
        if self._result:
            raise PviError(self._result, self)        

    def __del__(self):
        self.kill

    def kill(self):
        '''
        PviObject.kill: kills this object
        this should be called when object is not beeing used anymore
        to save PVI resources
        '''
        if self._linkID != 0:
            self._pviConnection._linkIDs.pop(self._linkID) # remove from linkIDs
            self._pviConnection._pviObjects.remove(self) # remove from PviObjects
            self._result = PviUnlink(self._linkID)
            self._linkID = 0
            if self._result != 0:
                raise PviError(self._result, self)

    


# basics1.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
# and shows reading and writing of basic data types
#
# PLC counterpart is a simple program running on ArSim 
#
# content of 'myProg.var':
# VAR
#     myUSINT : USINT := 0;
#     myUINT : UINT := 0;
#     mySTRING : STRING[80] := '';
#     myWSTRING : WSTRING[80] := "";
#     myTIME : TIME := T#0S;
#     myTOD : TIME_OF_DAY := TOD#00:00:00;
#     myDT : DATE_AND_TIME := DT#1970-01-01-00:00:00;
#     myDINT : DINT := 0;
#     myUDINT : UDINT := 0;
#     myDATE : DATE := D#1970-01-01;
#     myLREAL : LREAL := 0.0;
#     myREAL : REAL := 0.0;
#     myINT : INT := 0;
#     mySINT : SINT := 0;
#     myBOOL : BOOL := FALSE;
#     myINTArray : ARRAY[0..9] OF INT := [10(0)];
#     mySTRINGArray : ARRAY[0..9] OF STRING[80] := [10('')];
#     myREALArray : ARRAY[0..9] OF REAL := [10(0.0)];
# END_VAR
#
# content of 'myProg.st':
#   PROGRAM _INIT
#  	    myUSINT;
#  	    myUINT;
# 	    mySTRING;
# 	    myWSTRING;
# 	    myTIME;
# 	    myTOD;
# 	    myDT;
# 	    myDINT;
# 	    myINTArray;
# 	    mySTRINGArray;
#       myREALArray;
# 	    myUDINT;
# 	    myDATE;
# 	    myLREAL;
# 	    myREAL;
# 	    myINT;
# 	    mySINT;
# 	    myBOOL;
#   END_PROGRAM
#
#   PROGRAM _CYCLIC
#
#   END_PROGRAM
#
#

from time import sleep
from datetime import time, date, timedelta, datetime
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# line ANSL is the 'modern' way to access PLC variables
# (compared to the older INA2000 line)
#
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )
task1 = Task( cpu, 'myProg')

# create a list with variable names and values we want to write into
variableList = (   ('myBOOL', True), 
                ('myUSINT', 123),
                ('mySINT', -123),
                ('myUINT', 1234),
                ('myINT', -1234),
                ('myUDINT', 1234567),
                ('myDINT', -124567),                
                ('myREAL', 1.234e2 ),
                ('myLREAL', -3.1415e100),
                ('mySTRING', b'hello, world'),
                ('myWSTRING', 'hello, world'),
                ('myDT', datetime.fromisoformat('2011-11-04T00:05:23')),
                ('myTIME', timedelta(minutes = 3, seconds = 45, milliseconds= 33)),
                ('myTOD', time(hour=14, minute= 30, second = 13)),
                ('myDATE', date.today()),
                ('myINTArray', (1,2,3,4,5,6,7) ), # write the first seven elements
                ('mySTRINGArray', (b'Huey', b'Dewey', b'Louie') ), # write the first three elements
                ('myREALArray', 100.0) # set all elements to the same value
            )

run = True

def taskErrorChanged( error : int ):
    global run

    if error != 0:
        raise PviError(error)
    else: # connection to task available
        try:
            for name, value in variableList:
                variable = Variable( task1, name )
                variable.value = value # write into
                print( name + ' : ' + str(variable.value) ) # read from
                variable.kill
        except BaseException as e:
            print( e )
        finally:
            run = False


task1.errorChanged = taskErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)



# profiler2.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# extract all profiler data from a systemdump container and save it to csv
# a separate *.txt file is created with object information
#

import tarfile
import tempfile
import csv
import os
import re
import locale
from pprint import PrettyPrinter
from pvi.utils import BrProfilerDataFile

locale.setlocale(locale.LC_ALL, '')

pattern = re.compile(r'(BuR_SDM_profiling_.*pd)')

systemdump = r'C:\Temp\BuR_SDM_Sysdump_2023-07-31_11-11-41.tar.gz'

with tempfile.TemporaryDirectory() as t:
    tf = tarfile.open( systemdump, mode = 'r')
    tf.extractall(t)
    dataDirName = t + '/Data Files'
    dataFileNames = os.listdir(dataDirName)
    for dataFileName in dataFileNames:
        match = pattern.findall(dataFileName)
        if match:
            profilingName = match[0]
            try:
                profilerData = BrProfilerDataFile(dataDirName + '/' + dataFileName)
            except TypeError: # ingore all non-profiler files
                continue  

            with open(f'{dataFileName}.txt', 'w', newline = '') as f: 
                pp = PrettyPrinter( indent = 4, stream = f )
                print( dataFileName,  '\n\nSummary:\n', file = f)
                pp.pprint( profilerData.info )
                print( '\n\ntask classes:\n', file = f )
                pp.pprint( profilerData.taskClasses )                
                print( '\n\ncyclic tasks:\n', file = f )
                pp.pprint( profilerData.cyclicTasks )                
                print( '\n\nsystem tasks:\n', file = f )
                pp.pprint( profilerData.systemTasks )                
                print( '\n\ninterrupt handler:\n', file = f )
                pp.pprint( profilerData.interruptHandler )                
                print( '\n\nprofiled library functions:\n', file = f )
                pp.pprint( profilerData.libraryFunctions )                 
                
                
            with open(f'{dataFileName}.csv', 'w', newline = '') as f: 
                writer = csv.writer(f, delimiter= ';', quotechar="'" )
                writer.writerow ( ["Nr", "time", "object name", "ident","event", "event description"])            
                for n, e in enumerate(profilerData.entries):
                    writer.writerow( ( n+1, locale.format_string( '%.6f', e.microseconds), e.objectName, hex(e.objectIdent), hex(e.event), e.eventDescription ) ) 







#
# Pvi.py
# Python connector for B&R Pvi (process visualization interface)
#
#  https://github.com/hilch/Pvi.py
# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, publish, distribute, 
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from datetime import datetime, timedelta
from collections import namedtuple
import struct
from .BrFile import *

BrLoggerFileEntry = namedtuple('LoggerEntry', [ 'RecordID', 'Time', 'Nanosec', 'ObjectID' , 'Severity', 'Code', 
                         'EventID', 'OriginID', 'AsciiData', 'BinaryData'])

class BrLoggerFile(BrFile):
    '''
    class for a *.br file containing logger data
    '''
    def __init__(self, filename: str):
        super().__init__(filename)
        if self.fileType != BrFileType.LOGGER_MODULE:
            raise TypeError(f'content is not a B&R logger module (Type is {self.fileType})')
        self.__entries = None
        self.__BASE_OFFSET = 0xc0
        self.__logDataLength = struct.unpack_from('<I', self._content, 0x90)[0] # uint32
        self.__version = struct.unpack_from('<H', self._content, 0x94)[0] # uint16
        self.__actIndex = struct.unpack_from('<I', self._content, 0x9c)[0] # uint32
        self.__actOffset = struct.unpack_from('<I', self._content, 0xa0)[0] # uint32
        self.__writeOffset = struct.unpack_from('<I', self._content, 0xa4)[0] # uint32
        self.__refIndex = struct.unpack_from('<I', self._content, 0xa8)[0] # uint32
        self.__refOffset = struct.unpack_from('<I', self._content, 0xac)[0] # uint32
        self.__invalidLength = struct.unpack_from('<I', self._content, 0xb0)[0] # uint32
        self.__offsetUtc = struct.unpack_from('<h', self._content, 0xb4)[0] # int16
        self.__daylightSaving = struct.unpack_from('<H', self._content, 0xb6)[0] # uint16

    ## offset 0x80: b'83 B9 57 2A 9C F3 60 00 83 B9 57 2A 9C F3 60 00' ??

    @property
    def logDataLength(self) -> int:
        '''
        configured logger length (bytes)
        '''
        return self.__logDataLength


    @property
    def version(self) -> int:
        '''
        log format version
        '''
        return self.__version


    @property
    def offsetUtc(self) -> int:
        '''
        offset to UTC in minutes
        '''
        return self.__offsetUtc


    @property
    def dst(self) -> bool:
        '''
        daylight saving time active ?
        '''
        return bool(self.__daylightSaving)


    @property
    def numberOfEntries(self) -> int:
        '''
        number of logger entries
        '''
        return len(self.entries)


    @property
    def lowestRecordID(self) -> int:
       '''
       lowest ID contained in this logger
       '''
       return self.entries[-1].RecordID

    @property
    def highestRecordID(self) -> int:
       '''
       highest ID contained in this logger
       '''       
       return self.entries[0].RecordID


    def __readEntry(self) -> bytes:
        '''
        (internal) read and decode an entry
        '''              
        recordId, lengthOfPrevious, lengthOfEntry, lengthOfBinaryData = struct.unpack_from('<IIII', self._content[self.__entryOffset:self.__entryOffset+0x10], 0) # decode header fields
        if recordId != (self.__lastRecordId - 1):
            return None # end of logger
        self.__lastRecordId = recordId # save for next call
        entry = bytes(self._content[self.__entryOffset:self.__entryOffset+lengthOfEntry])
        # entry[0x10:0x14] ?
        info1, info2, time, nanosecond, severity, code, objectId, eventId, originId, binaryData = struct.unpack_from(f'<HHIIII36sii{lengthOfBinaryData}s', entry, 0x14)
        time = datetime.utcfromtimestamp( time )
        microsecond, nanosecond = divmod(nanosecond,1000)
        time += timedelta( microseconds = microsecond, minutes = self.__offsetUtc)
        if severity < 4:        
            severity = ('Success', 'Info', 'Warning', 'Error')[severity]
        else:
            severity = hex(severity)
        objectId = objectId[:objectId.find(b'\x00')].decode(encoding = 'utf-8', errors = 'ignore') 
        if code == 0 and eventId != 0: # new logger format ?
            code = eventId & 0xffff         
        asciiData = ''
        if info1 == 8 and info2 == 1: # binary data contains ASCII string -> remove trailing bytes
            asciiData = binaryData[:binaryData.find(b'\x00')].decode(encoding = 'utf-8', errors = 'ignore')
        elif info1 == 0 and info2 == 3: # 'old' format
            asciiData = binaryData[:binaryData.find(b'\x00')].decode(encoding = 'utf-8', errors = 'ignore')
            binaryData = binaryData[binaryData.find(b'\x00')+1:]

        # entry[-8:] = ??

        if self.__entryOffset == self.__BASE_OFFSET:
            self.__entryOffset = self.__BASE_OFFSET - lengthOfPrevious + self.logDataLength - self.__invalidLength
        elif (self.__entryOffset - lengthOfPrevious) < self.__BASE_OFFSET:
              self.__entryOffset = self.__entryOffset - lengthOfPrevious + self.logDataLength - self.__invalidLength
        else:  
            self.__entryOffset -= lengthOfPrevious # point to the previous entry

        return BrLoggerFileEntry(
                    RecordID = recordId,
                    Time = time,
                    Nanosec= nanosecond,
                    ObjectID = objectId,
                    Severity = severity,
                    EventID= eventId,
                    OriginID = originId,
                    Code= code,
                    AsciiData = asciiData,
                    BinaryData = binaryData
                ) 
        


    
    @property
    def entries(self) -> list:
        if self.__entries == None:
            self.__entries = list()
            self.__entryOffset = self.__actOffset + self.__BASE_OFFSET
            self.__lastRecordId = self.__actIndex + 1    
            while entry := self.__readEntry():  
                self.__entries.append( entry )
        return self.__entries


    @property
    def xmlHeader(self):
        header = bytearray()
        for x in self._content[0x39:]:
            if x != 0:
                header.append(x)
            else:
                break
        return header.decode('ascii')
            

    def __repr__(self) -> str:
        return f'BrLoggerFile V{self.version}, Entries: {self.numberOfEntries}, Length {self.logDataLength}, RecordId: {self.lowestRecordID} to {self.highestRecordID}'            
    


# modules1.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on ArSim 
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this simple example creates a module on CPU by downloading a bytestream and checks if it exists
#


from time import sleep
from tkinter import FALSE
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# line ANSL is the 'modern' way to access PLC variables
# (compared to the older INA2000 line)
#
line = Line( pviConnection.root, 'LNANSL', CD='LNANSL')
device = Device( line, 'TCP', CD='/IF=TcpIp' )
cpu = Cpu( device, 'myArsim', CD='/IP=127.0.0.1' )

run = True
    
    
def callback_progress(percent):
    print( f'progress: {percent} %\r', end="")   


def callback_downloaded():
    global run
    modules = cpu.modules
    if 'bigmod' in modules:
        print(" 'bigmod' was sucessfully downloaded !")
    else:
        print("error: 'bigmod' not found !")
    run = False # exit


def cpuErrorChanged( error : int ):
    if error == 0:
        s = cpu.status.get("RunState")
        if s == 'RUN' or s == 'SERV': # download is only allowed in these modes
            step = 1
            cpu.downloadModule( bytes(5000000), MN='bigmod', MT='BRT', progress = callback_progress , downloaded = callback_downloaded )
            print("downloading...")
    else:
        raise PviError( error )
    
cpu.errorChanged = cpuErrorChanged

while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)







# list_objects5.py
# An example that shows the application possibilities of https://github.com/hilch/Pvi.py
#
# PLC counterpart is B&R's 'coffee machine' running on PP65 with AR 3.x
# as provided with Automation Studio 4.1.17.113 (which
# can be downloaded free of charge from https://www.br-automation.com)
#
# this example lists global and local variables and their content
#
# This is similar to list_objects2.py but since we use a control running < AR 4.x
# ANSL is not available here and we change to good old INA2000


from time import sleep
from pprint import pprint
from pvi import *

pviConnection = Connection() # start a Pvi connection

# all PVI objects must be registered hierarchically
# ANSL is not available and we use INA2000
#
line = Line( pviConnection.root, 'LNINA', CD='LNINA2')
device = Device( line, 'TCP', CD='/IF=TcpIp /SA=113' ) # always use a unique node number for PVI client even though you won't use node numbers
cpu = Cpu( device, 'myPP65', CD='/DAIP=10.49.40.222' )

# alternative: use 'INA node number' instead of IP address. INA node numbers must be unique in network !
# cpu = Cpu( device, 'myPP65', CD='/DA=32' )


run = True

def cpuErrorChanged( error : int):
    global run

    if error == 0:

        tasks = dict()
        for taskName in cpu.tasks:
            task = Task( cpu, taskName )
            variables = list()                    
            for variableName in task.variables:
                variable = Variable(task, variableName)
                try:
                    dataType = variable.dataType
                    value = variable.value
                except PviError as e:
                    print(e)
                variables.append( { "Name" : variableName, "Type" : dataType, "Value" : value} )
                variable.kill()
            tasks.update({ taskName : variables })
            task.kill
        # write content to file
        with open('content.txt', 'w') as f: 
            pprint(tasks, stream=f, indent = 4)

        variables = list()
        for variableName in cpu.variables:
            variable = Variable(cpu, variableName)
            try:
                dataType = variable.dataType
                value = variable.value
            except PviError as e:
                print(e)
            variables.append( { "Name" : variableName, "Type" : dataType, "Value" : value} )
            variable.kill()


        # append the content to existing file
        with open('content.txt', 'a') as f: 
            f.write("\nglobalVars =\n")
            pprint( variables, stream=f) 

        print('content.txt was created !')                      
        run = False
    
    elif error:
        raise PviError(error)


cpu.errorChanged = cpuErrorChanged


while run:
    pviConnection.doEvents() # must be cyclically called
    sleep(0.1)





